package domain

type RecipeRepository interface {
	CreateRecipe(recipe *Recipe) error
	GetRecipe(id int, userId *int) (*Recipe, error)
	GetRecipes(ids []int, userId *int) ([]Recipe, error)
	SearchRecipes(filters SearchFilters, userId *int, favorites bool) ([]Recipe, error)
	CreateRecipeTags(recipe Recipe, tags []string) error
	GetUserRecipes(id int) ([]Recipe, error)
	GetUserFavoriteRecipes(id int) ([]Recipe, error)
	GetRecipeTags(recipe *Recipe) error
	GetRecipeFavorite(recipe *Recipe, userId int) error
	GetRecipeOfTheWeek(userId *int) (*Recipe, error)
}
